from __future__ import annotations

import asyncio
import time
from dataclasses import dataclass
from typing import Dict, Optional

from aiohttp import ClientSession
from aiohttp.client_exceptions import ClientResponseError

from ...utils.json import JSON


class BossAuthenticationError(RuntimeError):
    pass


@dataclass(frozen=True)
class TokenKey:
    domain: str
    username: Optional[str] = None
    password: Optional[str] = None
    token: Optional[str] = None

    def __post_init__(self) -> None:
        if self.token is None:
            assert self.username is not None
            assert self.password is not None
        else:
            assert self.username is None
            assert self.password is None


class TokenRepository:
    def __init__(self, http_client: ClientSession):
        self.http_client = http_client
        self.token_infos: Dict[TokenKey, JSON] = {}

        self.client_id = "endpoint"
        self.login_realm = "boss"

    def _openid_url(self, key: TokenKey) -> str:
        # domain:   https://api.boss.neurodata.io
        # auth_url: https://auth.boss.neurodata.io/auth
        protocol = key.domain.split("://", 1)[0]
        domain_end = key.domain.split(".", 1)[1]
        auth_url = f"{protocol}://auth.{domain_end}/auth"

        return f"{auth_url}/realms/{self.login_realm}/protocol/openid-connect/"

    async def get(self, key: TokenKey) -> str:
        if key.token is not None:
            return "Token " + key.token

        request_new = True
        if key in self.token_infos:
            token_info = self.token_infos[key]
            if time.monotonic() - token_info["time"] < token_info["expires_in"] - 10:
                request_new = False

        if request_new:
            url = self._openid_url(key) + "token"
            data = {
                "grant_type": "password",
                "client_id": self.client_id,
                "username": key.username,
                "password": key.password,
            }
            now = time.monotonic()
            try:
                async with await self.http_client.post(url, data=data) as r:
                    token_info = await r.json()
            except ClientResponseError as e:
                if e.status == 401:  # Unauthorized
                    raise BossAuthenticationError(
                        f'Could not authorize user "{key.username}" at {key.domain}.'
                    )
                else:
                    raise e
            # {
            #     "access_token": "…",
            #     "expires_in": 1800,
            #     "refresh_expires_in": 3600,
            #     "refresh_token": "…",
            #     "token_type": "bearer",
            #     "id_token": "…",
            #     "not-before-policy": 0,
            #     "session_state": "…"
            # }

            token_info["time"] = now
            assert token_info["token_type"] == "bearer"
            self.token_infos[key] = token_info

        return "Bearer " + token_info["access_token"]

    async def get_header(self, key: TokenKey) -> Dict[str, str]:
        return {"Authorization": await self.get(key)}

    async def logout(self, key: TokenKey) -> None:
        if key.token is not None:
            return
        url = self._openid_url(key) + "logout"
        data = {
            "refresh_token": self.token_infos[key]["refresh_token"],
            "client_id": self.client_id,
        }
        try:
            await self.http_client.post(url, data=data)
            del self.token_infos[key]
        except ClientResponseError:
            pass

    async def logout_all(self) -> None:
        await asyncio.gather(
            *(self.logout(token_key) for token_key in self.token_infos)
        )



def convert_si_units(val: int, unit_in: str, unit_out: str) -> int:
    prefix_to_exponent = {
        "yotta": 24,
        "zetta": 21,
        "exa": 18,
        "peta": 15,
        "tera": 12,
        "giga": 9,
        "mega": 6,
        "kilo": 3,
        "hecto": 2,
        "deca": 1,
        "deci": -1,
        "centi": -2,
        "milli": -3,
        "micro": -6,
        "nano": -9,
        "pico": -12,
        "femto": -15,
        "atto": -18,
        "zepto": -21,
        "yocto": -24,
    }

    def get_exponent(unit: str) -> int:
        possible_exponent = [
            exponent
            for prefix, exponent in prefix_to_exponent.items()
            if unit.startswith(prefix)
        ]
        if len(possible_exponent) == 0:
            return 0
        return possible_exponent[0]

    return val * 10 ** (get_exponent(unit_in) - get_exponent(unit_out))



from pathlib import Path
from typing import Any, Dict

from setuptools_rust import RustExtension


def build(setup_kwargs: Dict[str, Any]) -> None:
    """
    This `build` function is mandatory in order to build the extensions, since poetry imports it from this module.
    """
    setup_kwargs.update(
        {
            "rust_extensions": [
                RustExtension(
                    "wkconnect.fast_wkw",
                    path=(Path("fast_wkw") / "Cargo.toml"),
                    quiet=False,
                )
            ],
            "zip_safe": False,
        }
    )



from dataclasses import dataclass
from functools import lru_cache
from typing import Optional, Tuple

import numpy as np
from async_lru import alru_cache
from wkcuber.api.Dataset import WKDataset
from wkcuber.api.Properties.LayerProperties import LayerProperties
from wkcuber.mag import Mag

from wkconnect.utils.types import Vec3D, Vec3Df

from ...fast_wkw import DatasetCache, DatasetHandle  # pylint: disable=no-name-in-module
from ...webknossos.models import BoundingBox as WkBoundingBox
from ...webknossos.models import DataLayer as WkDataLayer
from ...webknossos.models import DataSource as WkDataSource
from ...webknossos.models import DataSourceId as WkDataSourceId
from ..backend import DatasetInfo


@dataclass(frozen=True)
class Dataset(DatasetInfo):
    organization_name: str
    dataset_name: str
    dataset_handle: WKDataset
    wkw_cache: DatasetCache

    def to_webknossos(self) -> WkDataSource:
        return WkDataSource(
            WkDataSourceId(self.organization_name, self.dataset_name),
            [
                self.layer_to_webknossos(layer_name, layer)
                for layer_name, layer in self.dataset_handle.properties.data_layers.items()
            ],
            Vec3Df(*self.dataset_handle.properties.scale),
        )

    def layer_to_webknossos(
        self, layer_name: str, layer_properties: LayerProperties
    ) -> WkDataLayer:
        return WkDataLayer(
            layer_name,
            layer_properties._category,
            WkBoundingBox(
                topLeft=Vec3D(*layer_properties._bounding_box["topLeft"]),
                width=layer_properties._bounding_box["width"],
                height=layer_properties._bounding_box["height"],
                depth=layer_properties._bounding_box["depth"],
            ),
            [
                Vec3D(*mag.mag.to_array())
                for mag in layer_properties._wkw_magnifications
            ],
            layer_properties._element_class,
        )

    @lru_cache(maxsize=1000)
    def get_data_handle(
        self, layer_name: str, zoom_step: int
    ) -> Optional[Tuple[DatasetHandle, Mag]]:
        layer = self.dataset_handle.get_layer(layer_name)
        # Finding the right mag for the zoom_step:
        # 2 ** zoomStep = max(mag.x, mag.y, mag.z)
        mag_datasets = [
            mag_dataset
            for mag_dataset in layer.mags.values()
            if (2 ** zoom_step) == Mag(mag_dataset.name).as_np().max()
        ]
        if len(mag_datasets) < 1:
            return None
        mag_dataset = mag_datasets[0]

        data_handle = self.wkw_cache.get_dataset(str(mag_dataset.view.path))
        return (data_handle, Mag(mag_dataset.name))

    @alru_cache(maxsize=2 ** 12, cache_exceptions=False)
    async def read_data(
        self, layer_name: str, zoom_step: int, wk_offset: Vec3D, shape: Vec3D
    ) -> Optional[np.ndarray]:
        assert shape == Vec3D(
            32, 32, 32
        ), "Only buckets of 32 edge length are supported"
        assert shape % 32 == Vec3D(0, 0, 0), "Only 32-aligned buckets are supported"
        data_handle_opt = self.get_data_handle(layer_name, zoom_step)
        if data_handle_opt is None:
            return None
        data_handle, mag = data_handle_opt
        offset = (
            np.array([wk_offset.x, wk_offset.y, wk_offset.z]) / mag.as_np()
        ).astype(np.uint32)
        block = await data_handle.read_block(tuple(offset))
        return np.frombuffer(block.buf, dtype=np.dtype(block.dtype)).reshape(
            block.shape, order="F"
        )

    def clear_cache(self) -> None:
        self.read_data.cache_clear()  # pylint: disable=no-member
        self.get_data_handle.cache_clear()  # pylint: disable=no-member
        self.wkw_cache.clear_cache_prefix(str(self.dataset_handle.path))



from typing import List, Union

from aiohttp import ClientSession
from sanic.config import Config

from ..utils.caching import atlru_cache
from ..utils.json import from_json, to_json
from .access import AccessAnswer, AccessRequest
from .models import DataSource, DataStoreStatus, UnusableDataSource


class WebKnossosClient:
    def __init__(self, config: Config, http_client: ClientSession) -> None:
        self.webknossos_url = config["webknossos"]["url"]
        self.datastore_url = config["server"]["url"]
        self.datastore_name = config["datastore"]["name"]
        self.datastore_key = config["datastore"]["key"]

        self.http_client = http_client

        self.headers = {"Content-Type": "application/json"}
        self.params = {"key": self.datastore_key}

    async def report_status(self, ok: bool = True) -> None:
        url = f"{self.webknossos_url}/api/datastores/{self.datastore_name}/status"
        status = DataStoreStatus(ok, self.datastore_url)
        await self.http_client.patch(
            url, headers=self.headers, params=self.params, json=to_json(status)
        )

    async def report_dataset(
        self, dataset: Union[DataSource, UnusableDataSource]
    ) -> None:
        url = f"{self.webknossos_url}/api/datastores/{self.datastore_name}/datasource"
        await self.http_client.put(
            url, headers=self.headers, params=self.params, json=to_json(dataset)
        )

    async def report_all_datasets(
        self, datasets: List[Union[DataSource, UnusableDataSource]]
    ) -> None:
        url = f"{self.webknossos_url}/api/datastores/{self.datastore_name}/datasources"
        await self.http_client.put(
            url, headers=self.headers, params=self.params, json=to_json(datasets)
        )

    @atlru_cache(seconds_to_use=2 * 60)
    async def request_access(
        self, token: str, access_request: AccessRequest
    ) -> AccessAnswer:
        url = f"{self.webknossos_url}/api/datastores/{self.datastore_name}/validateUserAccess"
        params = {"token": token, **self.params}
        async with await self.http_client.post(
            url,
            headers=self.headers,
            params=params,
            json=to_json(access_request),
            raise_for_status=False,
        ) as r:
            return from_json(await r.json(), AccessAnswer)



import asyncio
import logging
from typing import Dict, Optional, Tuple, cast

import blosc
import numpy as np
from aiohttp import ClientSession
from aiohttp.client_exceptions import ClientResponseError
from async_lru import alru_cache

from ...utils.si import convert_si_units
from ...utils.types import JSON, Box3D, HashableDict, Vec3D, Vec3Df
from ..backend import Backend, DatasetInfo
from .client import Client
from .models import Channel, Dataset, Experiment
from .token_repository import TokenKey, TokenRepository

logger = logging.getLogger()


class Boss(Backend):
    def __init__(self, config: Dict, http_client: ClientSession) -> None:
        self.tokens = TokenRepository(http_client)
        self.http_client = http_client
        self.client = Client(http_client, self.tokens)

    async def __handle_new_channel(
        self,
        domain: str,
        collection: str,
        experiment: str,
        channel: str,
        token_key: TokenKey,
    ) -> Tuple[str, Channel]:
        channel_info = await self.client.get_channel(
            domain, collection, experiment, channel, token_key
        )
        assert channel_info["base_resolution"] == 0
        if channel_info["downsample_status"] != "DOWNSAMPLED":
            logger.warn(
                f"BOSS did not finish downsampling for \"{'/'.join([domain, collection, experiment])}\", current status is {channel_info['downsample_status']}."
            )

        downsample_info = await self.client.get_downsample(
            domain, collection, experiment, channel, token_key
        )
        resolutions = tuple(
            Vec3Df(*i) for i in sorted(downsample_info["voxel_size"].values())
        )
        normalized_resolutions = tuple(
            (i / resolutions[0]).to_int() for i in resolutions
        )

        return (
            channel,
            Channel(
                channel_info["datatype"], normalized_resolutions, channel_info["type"]
            ),
        )

    async def handle_new_dataset(
        self, organization_name: str, dataset_name: str, dataset_info: JSON
    ) -> DatasetInfo:
        domain = dataset_info["domain"]
        assert domain.startswith("https://api") or domain.startswith("http://api")

        collection = dataset_info["collection"]
        experiment = dataset_info["experiment"]

        token_key = TokenKey(
            domain,
            dataset_info.get("username", None),
            dataset_info.get("password", None),
            dataset_info.get("token", None),
        )

        experiment_info = await self.client.get_experiment(
            domain, collection, experiment, token_key
        )
        assert experiment_info["num_time_samples"] == 1

        coord_info = await self.client.get_coord(
            domain, experiment_info["coord_frame"], token_key
        )
        bounding_box = Box3D(
            Vec3D(coord_info["x_start"], coord_info["y_start"], coord_info["z_start"]),
            Vec3D(coord_info["x_stop"], coord_info["y_stop"], coord_info["z_stop"]),
        )
        global_scale = Vec3Df(
            *(
                convert_si_units(
                    coord_info[f"{dim}_voxel_size"],
                    coord_info["voxel_unit"],
                    "nanometers",
                )
                for dim in "xyz"
            )
        )

        channels = await asyncio.gather(
            *(
                self.__handle_new_channel(
                    domain, collection, experiment, channel, token_key
                )
                for channel in experiment_info["channels"]
            )
        )

        experiment = Experiment(collection, experiment, HashableDict(channels))

        return Dataset(
            organization_name,
            dataset_name,
            domain,
            experiment,
            bounding_box,
            global_scale,
            token_key,
        )

    @alru_cache(maxsize=2 ** 12, cache_exceptions=False)
    async def __cached_read_data(
        self,
        abstract_dataset: DatasetInfo,
        layer_name: str,
        zoom_step: int,
        wk_offset: Vec3D,
        shape: Vec3D,
    ) -> Optional[np.ndarray]:
        dataset = cast(Dataset, abstract_dataset)
        channel = dataset.experiment.channels[layer_name]

        channel_offset = wk_offset // channel.resolutions[zoom_step]
        box = Box3D.from_size(channel_offset, shape)

        compressed_data = await self.client.get_cutout(
            dataset.domain,
            dataset.experiment.collection_name,
            dataset.experiment.experiment_name,
            layer_name,
            zoom_step,
            box,
            dataset.token_key,
        )

        byte_data = blosc.decompress(compressed_data)
        if channel.type == "color" and channel.datatype == "uint16":
            # this will be downscaled to uint8,
            # we want to take the higher-order bits here
            data = np.frombuffer(byte_data, dtype=">u2")
        else:
            data = np.frombuffer(byte_data, dtype=channel.datatype)
        return data.astype(channel.wk_datatype()).reshape(shape, order="F")

    async def read_data(
        self,
        abstract_dataset: DatasetInfo,
        layer_name: str,
        zoom_step: int,
        wk_offset: Vec3D,
        shape: Vec3D,
    ) -> Optional[np.ndarray]:
        try:
            return await self.__cached_read_data(
                abstract_dataset, layer_name, zoom_step, wk_offset, shape
            )
        except ClientResponseError:
            # will be reported in MISSING-BUCKETS, frontend will retry
            return None

    def clear_dataset_cache(self, dataset: DatasetInfo) -> None:
        pass

    async def on_shutdown(self) -> None:
        await self.tokens.logout_all()



import asyncio
import os

from sanic import Blueprint, response
from sanic.request import Request

trace = Blueprint(__name__, url_prefix="/triggers")

trace.static("/trace", "data/trace.html", name="trace_get")
trace.static("/trace/flame.svg", "data/flame.svg", name="trace_get")


@trace.route("/trace", methods=["POST"])
async def trace_post(request: Request) -> response.HTTPResponse:
    # py-spy is only in dev-packages
    p = await asyncio.create_subprocess_exec(
        "py-spy",
        "--flame",
        "data/flame.svg",
        "--duration",
        "20",
        "--pid",
        str(os.getpid()),
    )
    await p.wait()
    return (
        response.text("Ok") if p.returncode == 0 else response.text("Error", status=500)
    )



from aiohttp import ClientResponse, ClientSession

from ...utils.types import JSON, Box3D
from .token_repository import TokenKey, TokenRepository


class Client:
    def __init__(self, http_client: ClientSession, tokens: TokenRepository) -> None:
        self.http_client = http_client
        self.tokens = tokens

    async def auth_get(self, url: str, token_key: TokenKey) -> ClientResponse:
        return await self.http_client.get(
            url, headers=await self.tokens.get_header(token_key)
        )

    async def get_experiment(
        self, domain: str, collection: str, experiment: str, token_key: TokenKey
    ) -> JSON:
        # {
        #     "channels": ["annotation_50um", "average_50um", "nissl_50um"],
        #     "name": "sagittal_50um",
        #     "description": "Allen Reference Atlases ingested from the source at 50 um resolution",
        #     "collection": "ara_2016",
        #     "coord_frame": "ara_2016_50um",
        #     "num_hierarchy_levels": 1,
        #     "hierarchy_method": "isotropic",
        #     "num_time_samples": 1,
        #     "time_step": None,
        #     "time_step_unit": "",
        #     "creator": "vikramc"
        # }
        experiment_url = "/".join(
            [domain, "v1", "collection", collection, "experiment", experiment]
        )
        async with await self.auth_get(experiment_url, token_key) as r:
            return await r.json()

    async def get_coord(
        self, domain: str, coord_frame: str, token_key: TokenKey
    ) -> JSON:
        # {
        #     "name": "ara_2016_50um",
        #     "description": "50 um Allen reference atlas",
        #     "x_start": 0,
        #     "x_stop": 264,
        #     "y_start": 0,
        #     "y_stop": 160,
        #     "z_start": 0,
        #     "z_stop": 228,
        #     "x_voxel_size": 50.0,
        #     "y_voxel_size": 50.0,
        #     "z_voxel_size": 50.0,
        #     "voxel_unit": "micrometers"
        # }
        coord_url = f"{domain}/v1/coord/{coord_frame}"
        async with await self.auth_get(coord_url, token_key) as r:
            return await r.json()

    async def get_channel(
        self,
        domain: str,
        collection: str,
        experiment: str,
        channel: str,
        token_key: TokenKey,
    ) -> JSON:
        # {
        #     "name": "nissl_50um",
        #     "description": "",
        #     "experiment": "sagittal_50um",
        #     "default_time_sample": 0,
        #     "type": "image",                   | "annotation"
        #     "base_resolution": 0,
        #     "datatype": "uint16",
        #     "creator": "vikramc",
        #     "sources": [],
        #     "downsample_status": "DOWNSAMPLED",
        #     "related": []
        # }
        channel_url = "/".join(
            [
                domain,
                "v1",
                "collection",
                collection,
                "experiment",
                experiment,
                "channel",
                channel,
            ]
        )
        async with await self.auth_get(channel_url, token_key) as r:
            return await r.json()

    async def get_downsample(
        self,
        domain: str,
        collection: str,
        experiment: str,
        channel: str,
        token_key: TokenKey,
    ) -> JSON:
        # {
        #     "num_hierarchy_levels": 1,
        #     "cuboid_size": { "0": [512, 512, 16] },
        #     "extent": { "0": [264, 160, 228] },
        #     "status": "DOWNSAMPLED",
        #     "voxel_size": { "0": [50.0, 50.0, 50.0] }
        # }
        downsample_url = "/".join(
            [domain, "v1", "downsample", collection, experiment, channel]
        )
        async with await self.auth_get(downsample_url, token_key) as r:
            return await r.json()

    async def get_cutout(
        self,
        domain: str,
        collection: str,
        experiment: str,
        channel: str,
        resolution: int,
        range_box: Box3D,
        token_key: TokenKey,
    ) -> bytes:

        url = "/".join(
            [
                domain,
                "v1",
                "cutout",
                collection,
                experiment,
                channel,
                str(resolution),
                *(f"{left}:{right}" for left, right in zip(*range_box)),
            ]
        )
        async with await self.auth_get(url, token_key) as r:
            return await r.read()



from sanic import Blueprint

from .datasource_properties import datasource_properties
from .histogram import histogram
from .read_data import read_data
from .thumbnail import thumbnail
from .upload import upload

datasets = Blueprint.group(
    datasource_properties,
    read_data,
    thumbnail,
    histogram,
    upload,
    url_prefix="/datasets",
)



from typing import Dict, Tuple

from .backends.backend import DatasetInfo


class Repository:
    def __init__(self) -> None:
        self.datasets: Dict[Tuple[str, str], Tuple[str, DatasetInfo]] = {}

    def add_dataset(self, backend_name: str, dataset: DatasetInfo) -> None:
        self.datasets[(dataset.organization_name, dataset.dataset_name)] = (
            backend_name,
            dataset,
        )

    def get_dataset(
        self, organization_name: str, dataset_name: str
    ) -> Tuple[str, DatasetInfo]:
        return self.datasets[(organization_name, dataset_name)]



import logging
import math
from dataclasses import dataclass
from functools import lru_cache
from pathlib import Path
from typing import List, Optional, Tuple

import numpy as np
from tifffile import TiffFile, imread

from wkconnect.utils.types import Vec3D, Vec3Df

from ...webknossos.models import BoundingBox as WkBoundingBox
from ...webknossos.models import DataLayer as WkDataLayer
from ...webknossos.models import DataSource as WkDataSource
from ...webknossos.models import DataSourceId as WkDataSourceId
from ..backend import DatasetInfo

logger = logging.getLogger()

# Accessing the tiff header tags via their numbers:
TAG_TILE_WIDTH = 322
TAG_TILE_HEIGHT = 323
TAG_TILE_BYTE_OFFSETS = 324
TAG_PAGE_WIDTH = 256
TAG_PAGE_HEIGHT = 257


@dataclass(frozen=True)
class Dataset(DatasetInfo):
    organization_name: str
    dataset_name: str
    scale: Vec3Df
    path: Path
    untiled_size_maximum_mp: int

    def to_webknossos(self) -> WkDataSource:
        return WkDataSource(
            WkDataSourceId(self.organization_name, self.dataset_name),
            [self.layer_to_webknossos(layer_name) for layer_name in self.layers()],
            self.scale,
        )

    def layers(self) -> List[str]:
        layers = []
        for filepath in self.path.iterdir():
            if filepath.suffix == ".tif":
                layers.append(filepath.stem)
        assert len(layers) > 0, f"No valid layers found for tif dataset at {self.path}"
        return layers

    def layer_to_webknossos(self, layer_name: str) -> WkDataLayer:
        dtype, mags, page_shapes, _, _ = self.read_properties(layer_name)

        return WkDataLayer(
            layer_name,
            "segmentation" if layer_name.startswith("segmentation") else "color",
            WkBoundingBox(
                topLeft=Vec3D(0, 0, 0),
                width=page_shapes[0][0],
                height=page_shapes[0][1],
                depth=1,
            ),
            mags,
            str(self.repair_dtype(dtype)),
        )

    @lru_cache(maxsize=2 ** 12)
    def read_data(
        self, layer_name: str, zoom_step: int, wk_offset: Vec3D, shape: Vec3D
    ) -> Optional[np.ndarray]:
        mag_factor = 2 ** zoom_step
        wk_offset_scaled = wk_offset // mag_factor

        _, _, _, tile_shapes, _ = self.read_properties(layer_name)
        if tile_shapes is None:
            page_data = self.read_whole_page(layer_name, zoom_step)
            cropout = page_data[
                wk_offset_scaled[0] : wk_offset_scaled[0] + shape.x,
                wk_offset_scaled[1] : wk_offset_scaled[1] + shape.y,
            ]
            # note that page_data is already transposed, so the cropout does not
            # have to be transposed again, differing from the else branch
        else:
            tile_coordinate_offset, tile_data = self.read_tile(
                layer_name,
                zoom_step,
                (wk_offset_scaled.x, wk_offset_scaled.y),
                (shape.x, shape.y),
            )
            left_in_tile = wk_offset_scaled[0] - tile_coordinate_offset[0]
            top_in_tile = wk_offset_scaled[1] - tile_coordinate_offset[1]
            right_in_tile = left_in_tile + shape.x
            bottom_in_tile = top_in_tile + shape.y

            cropout = tile_data[
                top_in_tile:bottom_in_tile, left_in_tile:right_in_tile
            ].transpose()

        bucket = np.zeros(shape, dtype=cropout.dtype)
        bucket[0 : cropout.shape[0], 0 : cropout.shape[1], 0] = cropout
        return bucket

    def layer_filepath(self, layer_name: str) -> Path:
        return self.path / f"{layer_name}.tif"

    def clear_cache(self) -> None:
        self.read_properties.cache_clear()  # pylint: disable=no-member
        self.read_mmapped.cache_clear()  # pylint: disable=no-member

    @staticmethod
    def is_tiled(tif: TiffFile) -> bool:
        return any(
            TAG_TILE_WIDTH in page.tags
            and TAG_TILE_HEIGHT in page.tags
            and TAG_TILE_BYTE_OFFSETS in page.tags
            for page in tif.pages
        )

    @lru_cache(5)
    def read_properties(
        self, layer_name: str
    ) -> Tuple[
        np.dtype,
        List[Vec3D],
        List[Tuple[int, int]],
        Optional[List[Tuple[int, int]]],
        Optional[List[List[int]]],
    ]:
        filepath = self.layer_filepath(layer_name)
        with TiffFile(str(filepath)) as tif:
            assert (
                len(tif.pages) > 0
            ), f"No magnifications found. Empty tiff at {str(filepath)}?"

            mags = [Vec3D(2 ** mag, 2 ** mag, 1) for mag in range(len(tif.pages))]
            page_shapes = [
                (page.tags[TAG_PAGE_WIDTH].value, page.tags[TAG_PAGE_HEIGHT].value)
                for page in tif.pages
            ]
            dtype = tif.byteorder + tif.pages[0].dtype.char

            if Dataset.is_tiled(tif):
                for i, page in enumerate(tif.pages):
                    assert (
                        TAG_TILE_WIDTH in page.tags
                        and TAG_TILE_HEIGHT in page.tags
                        and TAG_TILE_BYTE_OFFSETS in page.tags
                        and TAG_PAGE_WIDTH in page.tags
                        and TAG_PAGE_HEIGHT in page.tags
                    ), (
                        f"Incomplete tags in tif file at {str(filepath)} for page {i}. "
                        + "Only tiled tifs with resolution pyramid are supported."
                    )
                tile_byte_offsets = [
                    page.tags[TAG_TILE_BYTE_OFFSETS].value for page in tif.pages
                ]
                tile_shapes = [
                    (page.tags[TAG_TILE_WIDTH].value, page.tags[TAG_TILE_HEIGHT].value)
                    for page in tif.pages
                ]
                for i, tile_shape in enumerate(tile_shapes):
                    assert (
                        math.log2(tile_shape[0]).is_integer()
                        and math.log2(tile_shape[1]).is_integer()
                        and tile_shape[0] >= 32
                        and tile_shape[1] >= 32
                        and tile_shape[0] <= 2048
                        and tile_shape[1] <= 2048
                    ), f"Tiff tile shapes must be power of two, min 32 and max 2048. Found {tile_shape} in page {i} at {str(filepath)}."
                return dtype, mags, page_shapes, tile_shapes, tile_byte_offsets
            else:
                for i, page_shape in enumerate(page_shapes):
                    assert (
                        page_shape[0] * page_shape[1]
                        <= self.untiled_size_maximum_mp * 10 ** 6
                    ), f"Tiff file has {page_shape[0]}*{page_shape[1]} px in page {i} at {str(filepath)} without tiles. To support images with >{self.untiled_size_maximum_mp} MP, please create a tiled tif instead."
                return dtype, mags, page_shapes, None, None

    @lru_cache(10)
    def read_whole_page(self, layer_name: str, page: int) -> np.ndarray:
        return imread(self.layer_filepath(layer_name), key=page).transpose()

    def read_tile(
        self,
        layer_name: str,
        page: int,
        target_offset: Tuple[int, int],
        target_shape: Tuple[int, int],
    ) -> Tuple[Tuple[int, int], np.ndarray]:
        dtype, _, page_shapes, tile_shapes, tile_byte_offsets = self.read_properties(
            layer_name
        )
        assert (
            tile_shapes is not None and tile_byte_offsets is not None
        ), "read_tile called for non-tiled tif"

        tile_shape = tile_shapes[page]

        target_tile = (
            target_offset[0] // tile_shape[0],
            target_offset[1] // tile_shape[1],
        )
        target_tile_bottomright = (
            (target_offset[0] + target_shape[0] - 1) // tile_shape[0],
            (target_offset[1] + target_shape[1] - 1) // tile_shape[1],
        )
        assert target_tile == target_tile_bottomright, (
            f"Requested tif data at {target_offset} from page {page} of {str(self.path)} "
            + "that cannot be served by loading a single tile."
        )

        tile_coordinate_offset = (
            target_tile[0] * tile_shape[0],
            target_tile[1] * tile_shape[1],
        )

        tiles_per_row = math.ceil(page_shapes[page][0] / tile_shape[0])
        target_tile_index = target_tile[0] + target_tile[1] * tiles_per_row
        tile_byte_offset = tile_byte_offsets[page][target_tile_index]

        data = self.read_mmapped(layer_name, dtype, tile_byte_offset, tile_shape)
        return tile_coordinate_offset, data

    @lru_cache(maxsize=2 ** 9)
    def read_mmapped(
        self, layer_name: str, dtype: np.dtype, byte_offset: int, shape: Tuple[int, int]
    ) -> np.ndarray:
        image = np.zeros(shape, dtype=self.repair_dtype(dtype))
        mapped = np.memmap(
            str(self.layer_filepath(layer_name)), dtype, "r", byte_offset, shape, "C"
        )
        # copy content from mmap handle to array in order to avoid holding open thousands of handles.
        image[:] = mapped
        del mapped
        return image

    @lru_cache(5)
    def repair_dtype(self, dtype: np.dtype) -> np.dtype:
        # wk does not support signed datasets. We convert their data to the respective unsigned counterparts.
        dtype = np.dtype(dtype)
        if dtype == np.int8:
            return np.uint8(0).dtype
        if dtype == np.int16:
            return np.uint16(0).dtype
        if dtype == np.int32:
            return np.uint32(0).dtype
        return dtype



import asyncio
from functools import wraps
from typing import Any, Awaitable, Callable, TypeVar, cast

T = TypeVar("T", bound=Callable[..., Awaitable[Any]])


def repeat_every_seconds(
    interval_seconds: float, initial_call: bool = True
) -> Callable[[T], T]:
    def decorator(f: T) -> T:
        @wraps(f)
        async def decorator(*args: Any, **kwargs: Any) -> Any:
            if initial_call:
                await f(*args, **kwargs)
            while True:
                try:
                    await asyncio.sleep(interval_seconds)
                except asyncio.CancelledError:
                    break
                await f(*args, **kwargs)

        # cast necessary, see
        # https://github.com/python/mypy/issues/1927
        return cast(T, decorator)

    return decorator



from dataclasses import InitVar, dataclass, field
from typing import Any, Dict, Optional, Tuple

from gcloud.aio.auth import Token

from ...utils.types import Box3D, Vec3D
from ...webknossos.models import BoundingBox
from ...webknossos.models import DataLayer as WkDataLayer
from ...webknossos.models import DataSource as WkDataSource
from ...webknossos.models import DataSourceId as WkDataSourceId
from ..backend import DatasetInfo


@dataclass(frozen=True)
class Scale:
    chunk_sizes: Tuple[Vec3D, ...]
    encoding: str
    key: str
    resolution: Vec3D
    size: Vec3D
    voxel_offset: Vec3D
    compressed_segmentation_block_size: Optional[Vec3D] = None

    def __post_init__(self) -> None:
        assert len(self.chunk_sizes) > 0
        assert self.encoding in ["raw", "jpeg", "compressed_segmentation"]
        if self.encoding == "compressed_segmentation":
            assert self.compressed_segmentation_block_size is not None

    def box(self) -> Box3D:
        return Box3D.from_size(self.voxel_offset, self.size)

    def bounding_box(self) -> BoundingBox:
        return BoundingBox(self.voxel_offset, *self.size)


@dataclass(frozen=True)
class Layer:
    source: str
    data_type: str
    num_channels: int
    scales: Tuple[Scale, ...]
    relative_scale: Vec3D
    type: str
    largestSegmentId: Optional[int] = None
    # InitVar allows to consume mesh argument in init without storing it
    mesh: InitVar[Any] = None

    def __post_init__(self, mesh: Any) -> None:
        supported_data_types = {
            "image": ["uint8", "uint16", "uint32", "uint64"],
            "segmentation": ["uint32", "uint64"],
        }
        assert self.type in supported_data_types
        assert self.data_type in supported_data_types[self.type]
        assert self.num_channels == 1
        assert all(
            max(scale.resolution) == 2 ** i for i, scale in enumerate(self.scales)
        )

    def wk_data_type(self) -> str:
        if self.type == "segmentation":
            return "uint32"
        return self.data_type

    def to_webknossos(self, layer_name: str) -> WkDataLayer:
        return WkDataLayer(
            layer_name,
            {"image": "color", "segmentation": "segmentation"}[self.type],
            self.scales[0].bounding_box(),
            [scale.resolution for scale in self.scales],
            self.wk_data_type(),
            largestSegmentId=self.largestSegmentId,
        )


@dataclass
class Dataset(DatasetInfo):
    organization_name: str
    dataset_name: str
    layers: Dict[str, Layer]
    token: Optional[Token] = None
    scale: Vec3D = field(init=False)

    def __post_init__(self) -> None:
        relative_scale = set(layer.relative_scale for layer in self.layers.values())
        assert len(relative_scale) == 1
        self.scale = relative_scale.pop()

    def to_webknossos(self) -> WkDataSource:
        return WkDataSource(
            WkDataSourceId(self.organization_name, self.dataset_name),
            [
                layer.to_webknossos(layer_name)
                for layer_name, layer in self.layers.items()
            ],
            self.scale.to_float(),
        )



import logging
from pathlib import Path
from typing import Dict, Optional, cast

import numpy as np
from aiohttp import ClientSession
from wkcuber.api.Dataset import WKDataset

from ...fast_wkw import DatasetCache  # pylint: disable=no-name-in-module
from ...utils.types import JSON, Vec3D
from ..backend import Backend, DatasetInfo
from .models import Dataset

logger = logging.getLogger()


class Wkw(Backend):
    wkw_cache: DatasetCache

    def __init__(self, config: Dict, http_client: ClientSession) -> None:
        super().__init__(config, http_client)
        cache_size = config.get("fileCacheMaxEntries", 1000)
        self.wkw_cache = DatasetCache(cache_size)

    async def handle_new_dataset(
        self, organization_name: str, dataset_name: str, dataset_info: JSON
    ) -> DatasetInfo:

        path = Wkw.path(dataset_info, organization_name, dataset_name)
        return Dataset(
            organization_name, dataset_name, WKDataset(str(path)), self.wkw_cache
        )

    async def read_data(
        self,
        abstract_dataset: DatasetInfo,
        layer_name: str,
        zoom_step: int,
        wk_offset: Vec3D,
        shape: Vec3D,
    ) -> Optional[np.ndarray]:
        dataset = cast(Dataset, abstract_dataset)
        return await dataset.read_data(layer_name, zoom_step, wk_offset, shape)

    def clear_dataset_cache(self, abstract_dataset: DatasetInfo) -> None:
        dataset = cast(Dataset, abstract_dataset)
        dataset.clear_cache()

    @staticmethod
    def path(dataset_info: JSON, organization_name: str, dataset_name: str) -> Path:
        if "path" in dataset_info:
            return Path(dataset_info["path"])
        else:
            return Path("data", "binary", organization_name, dataset_name)



from dataclasses import dataclass
from typing import Dict, List, Tuple

from ...utils.types import Box3D, Vec3D, Vec3Df
from ...webknossos.models import BoundingBox as WkBoundingBox
from ...webknossos.models import DataLayer as WkDataLayer
from ...webknossos.models import DataSource as WkDataSource
from ...webknossos.models import DataSourceId as WkDataSourceId
from ..backend import DatasetInfo
from .token_repository import TokenKey


class BossResolutionsError(RuntimeError):
    pass


@dataclass(unsafe_hash=True)
class Channel:
    datatype: str
    resolutions: Tuple[Vec3D, ...]
    type: str

    def __post_init__(self) -> None:
        supported_datatypes = {
            "color": ["uint8", "uint16"],
            "segmentation": ["uint32", "uint64"],
        }
        if self.type == "image":
            self.type = "color"
        elif self.type == "annotation":
            self.type = "segmentation"
        assert self.type in supported_datatypes
        assert self.datatype in supported_datatypes[self.type]
        if not all(max(res) == 2 ** i for i, res in enumerate(self.resolutions)):
            error_msg = "\n".join(
                [
                    "Boss resolutions do not follow power of two factors:",
                    str(self.resolutions),
                    f"Maximum dimensions should be {[2 ** i for i in range(len(self.resolutions))]}.",
                ]
            )
            raise BossResolutionsError(error_msg)

    def wk_datatype(self) -> str:
        return "uint8" if self.type == "color" else "uint32"


@dataclass(frozen=True)
class Experiment:
    collection_name: str
    experiment_name: str
    channels: Dict[str, Channel]

    def webknossos_layers(self, wk_bounding_box: WkBoundingBox) -> List[WkDataLayer]:
        return [
            WkDataLayer(
                channel_name,
                channel.type,
                wk_bounding_box,
                list(channel.resolutions),
                channel.wk_datatype(),
            )
            for channel_name, channel in self.channels.items()
        ]


@dataclass(frozen=True)
class Dataset(DatasetInfo):
    organization_name: str
    dataset_name: str
    domain: str
    experiment: Experiment
    bounding_box: Box3D
    global_scale: Vec3Df
    token_key: TokenKey

    def to_webknossos(self) -> WkDataSource:
        bounding_box = WkBoundingBox.from_box(self.bounding_box)
        return WkDataSource(
            WkDataSourceId(self.organization_name, self.dataset_name),
            self.experiment.webknossos_layers(bounding_box),
            self.global_scale,
        )



from __future__ import annotations

from operator import add, floordiv, mod, mul, sub, truediv
from typing import Any, Callable, Iterator, NamedTuple, Union

import numpy as np


class Vec3D(NamedTuple):
    x: int
    y: int
    z: int

    def _element_wise(self, other: Any, fn: Callable[[int, Any], int]) -> Vec3D:
        if isinstance(other, tuple):
            return Vec3D(*(fn(a, b) for a, b in zip(self, other)))
        return Vec3D(*(fn(a, other) for a in self))  # pylint: disable=not-an-iterable

    def __add__(self, other: Any) -> Vec3D:
        return self._element_wise(other, add)

    def __sub__(self, other: Any) -> Vec3D:
        return self._element_wise(other, sub)

    def __mul__(self, other: Any) -> Vec3D:
        return self._element_wise(other, mul)

    def __floordiv__(self, other: Any) -> Vec3D:
        return self._element_wise(other, floordiv)

    def __mod__(self, other: Any) -> Vec3D:
        return self._element_wise(other, mod)

    def ceildiv(self, other: Any) -> Vec3D:
        return (self + other - 1) // other

    def pairmax(self, other: Any) -> Vec3D:
        return self._element_wise(other, max)

    def pairmin(self, other: Any) -> Vec3D:
        return self._element_wise(other, min)

    def to_float(self) -> Vec3Df:
        return Vec3Df(*map(float, self))

    def to_int(self) -> Vec3D:
        return Vec3D(*map(int, self))

    @classmethod
    def zeros(cls) -> Vec3D:
        return cls(0, 0, 0)

    @classmethod
    def ones(cls) -> Vec3D:
        return cls(1, 1, 1)


class Vec3Df(NamedTuple):
    x: float
    y: float
    z: float

    def _element_wise(self, other: Any, fn: Callable[[float, Any], float]) -> Vec3Df:
        if isinstance(other, tuple):
            return Vec3Df(*(fn(a, b) for a, b in zip(self, other)))
        return Vec3Df(*(fn(a, other) for a in self))  # pylint: disable=not-an-iterable

    def __add__(self, other: Any) -> Vec3Df:
        return self._element_wise(other, add)

    def __sub__(self, other: Any) -> Vec3Df:
        return self._element_wise(other, sub)

    def __mul__(self, other: Any) -> Vec3Df:
        return self._element_wise(other, mul)

    def __floordiv__(self, other: Any) -> Vec3Df:
        return self._element_wise(other, floordiv)

    def __truediv__(self, other: Any) -> Vec3Df:
        return self._element_wise(other, truediv)

    def __mod__(self, other: Any) -> Vec3Df:
        return self._element_wise(other, mod)

    def to_int(self) -> Vec3D:
        return Vec3D(*map(int, self))

    @classmethod
    def zeros(cls) -> Vec3Df:
        return cls(0.0, 0.0, 0.0)


class Box3D(NamedTuple):
    """
    3-dimensional Box [left, right)
    """

    left: Vec3D
    right: Vec3D

    @classmethod
    def from_size(cls, left: Vec3D, size: Vec3D) -> Box3D:
        return cls(left, left + size)

    def to_int(self) -> Box3D:
        return Box3D(self.left.to_int(), self.right.to_int())

    def size(self) -> Vec3D:
        return self.right - self.left

    def center(self) -> Vec3D:
        return self.left + self.size() // 2

    def _element_wise(
        self, other: Union[Vec3D, int], fn: Callable[[Vec3D, Union[Vec3D, int]], Vec3D]
    ) -> Box3D:
        return Box3D(*(fn(a, other) for a in self))  # pylint: disable=not-an-iterable

    def __add__(self, other: Any) -> Box3D:
        return self._element_wise(other, add)

    def __sub__(self, other: Any) -> Box3D:
        return self._element_wise(other, sub)

    def __mul__(self, other: Any) -> Box3D:
        return self._element_wise(other, mul)

    def div(self, other: Any) -> Box3D:
        return Box3D(self.left // other, self.right.ceildiv(other))

    def union(self, other: Box3D) -> Box3D:
        return Box3D(self.left.pairmin(other.left), self.right.pairmax(other.right))

    def intersect(self, other: Box3D) -> Box3D:
        return Box3D(self.left.pairmax(other.left), self.right.pairmin(other.right))

    def range(self, offset: Vec3D = Vec3D.ones()) -> Iterator[Vec3D]:
        for x in range(self.left.x, self.right.x, offset.x):
            for y in range(self.left.y, self.right.y, offset.y):
                for z in range(self.left.z, self.right.z, offset.z):
                    yield Vec3D(x, y, z)

    def np_slice(self) -> np.lib.index_tricks.IndexExpression:
        return np.index_exp[
            self.left.x : self.right.x,
            self.left.y : self.right.y,
            self.left.z : self.right.z,
        ]


# TODO refine this when recursive types are possible:
# see https://github.com/python/mypy/issues/731
JSON = Any


class HashableDict(dict):
    def __hash__(self) -> int:  # type: ignore  # dict returns None, we need int
        return hash(tuple(sorted(self.items())))



import logging
from pathlib import Path
from typing import Dict, Optional, cast

import numpy as np
from aiohttp import ClientSession

from ...utils.types import JSON, Vec3D, Vec3Df
from ..backend import Backend, DatasetInfo
from .models import Dataset

logger = logging.getLogger()


class Tiff(Backend):
    def __init__(self, config: Dict, http_client: ClientSession) -> None:
        super().__init__(config, http_client)
        self.config = config

    async def handle_new_dataset(
        self, organization_name: str, dataset_name: str, dataset_info: JSON
    ) -> DatasetInfo:
        scale = (
            Vec3Df(dataset_info["scale"][0], dataset_info["scale"][1], 1)
            if "scale" in dataset_info
            else Vec3Df(1, 1, 1)
        )
        path = Tiff.path(dataset_info, organization_name, dataset_name)
        untiled_size_maximum_mp = (
            int(self.config["tif_untiled_size_maximum_mp"])
            if "tif_untiled_size_maximum_mp" in self.config
            else 20
        )
        return Dataset(
            organization_name, dataset_name, scale, path, untiled_size_maximum_mp
        )

    async def read_data(
        self,
        abstract_dataset: DatasetInfo,
        layer_name: str,
        zoom_step: int,
        wk_offset: Vec3D,
        shape: Vec3D,
    ) -> Optional[np.ndarray]:
        dataset = cast(Dataset, abstract_dataset)
        return dataset.read_data(layer_name, zoom_step, wk_offset, shape)

    def clear_dataset_cache(self, abstract_dataset: DatasetInfo) -> None:
        dataset = cast(Dataset, abstract_dataset)
        dataset.clear_cache()

    @staticmethod
    def path(dataset_info: JSON, organization_name: str, dataset_name: str) -> Path:
        if "path" in dataset_info:
            return Path(dataset_info["path"])
        else:
            return Path("data", "binary", organization_name, dataset_name)



import base64
from io import BytesIO

from PIL import Image
from sanic import Blueprint, response
from sanic.request import Request

from ...utils.colors import color_bytes
from ...utils.types import Vec3D
from ...webknossos.access import AccessRequest, authorized

thumbnail = Blueprint(__name__)


@thumbnail.route(
    "/<organization_name>/<dataset_name>/layers/<layer_name>/thumbnail.json"
)
@authorized(AccessRequest.read_dataset)
async def get_thumbnail(
    request: Request, organization_name: str, dataset_name: str, layer_name: str
) -> response.HTTPResponse:
    width = int(request.args.get("width"))
    height = int(request.args.get("height"))

    (backend_name, dataset) = request.app.ctx.repository.get_dataset(
        organization_name, dataset_name
    )
    backend = request.app.ctx.backends[backend_name]
    layer = [i for i in dataset.to_webknossos().dataLayers if i.name == layer_name][0]
    scale = min(3, len(layer.resolutions) - 1)
    center = layer.boundingBox.box().center()
    size = Vec3D(width, height, 1)
    data = (
        await backend.read_data(dataset, layer_name, scale, center - size // 2, size)
    )[:, :, 0]
    if layer.category == "segmentation":
        data = data.astype("uint8")
        thumbnail = Image.fromarray(data, mode="P")
        color_list = list(color_bytes.values())[: 2 ** 8]
        thumbnail.putpalette(b"".join(color_list))
        with BytesIO() as output:
            thumbnail.save(output, "PNG", transparency=0)
            return response.json(
                {"mimeType": "image/png", "value": base64.b64encode(output.getvalue())}
            )
    else:
        thumbnail = Image.fromarray(data)
        with BytesIO() as output:
            thumbnail.save(output, "JPEG")
            return response.json(
                {"mimeType": "image/jpeg", "value": base64.b64encode(output.getvalue())}
            )



from typing import TYPE_CHECKING, cast

from sanic import Blueprint, response
from sanic.request import Request

from ..webknossos.access import AccessRequest, authorized

if TYPE_CHECKING:
    from ..__main__ import Server

triggers = Blueprint(__name__, url_prefix="/triggers")


@triggers.route("/checkInboxBlocking")
@authorized(AccessRequest.administrate_datasets)
async def check_inbox_blocking(request: Request) -> response.HTTPResponse:
    app = cast("Server", request.app)
    await app.load_persisted_datasets()
    return response.text("Ok")


@triggers.route("/newOrganizationFolder")
@authorized(AccessRequest.administrate_datasets)
async def new_organization_folder(request: Request) -> response.HTTPResponse:
    return response.text("Ok")


@triggers.route("/checkNewOrganizationFolder")
async def check_new_organization_folder(request: Request) -> response.HTTPResponse:
    return response.text("Ok")


@triggers.route("/reload/<organization_name>/<dataset_name>")
@authorized(AccessRequest.administrate_datasets)
async def reload_dataset(
    request: Request, organization_name: str, dataset_name: str
) -> response.HTTPResponse:
    app = cast("Server", request.app)
    await app.load_persisted_datasets()
    (backend_name, dataset) = request.app.ctx.repository.get_dataset(
        organization_name, dataset_name
    )
    backend = app.ctx.backends[backend_name]
    backend.clear_dataset_cache(dataset)
    return response.text("Ok")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['wkconnect',
 'wkconnect.backends',
 'wkconnect.backends.boss',
 'wkconnect.backends.neuroglancer',
 'wkconnect.backends.tiff',
 'wkconnect.backends.wkw',
 'wkconnect.routes',
 'wkconnect.routes.datasets',
 'wkconnect.utils',
 'wkconnect.webknossos']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=6.2,<7.0',
 'aiohttp>=3.7,<4.0',
 'async-lru>=1.0,<2.0',
 'blosc>=1.10,<2.0',
 'brotlipy>=0.7.0,<0.8.0',
 'compressed_segmentation>=1.0,<2.0',
 'gcloud-aio-auth>=3.0,<4.0',
 'jpeg4py>=0.1.4,<0.2.0',
 'numpy>=1.17,<2.0',
 'sanic==21.3.4',
 'sanic_cors>=1.0.0,<2.0.0',
 'tifffile>=2020.9.3,<2021.0.0',
 'wkcuber>=0.5,<0.6']

setup_kwargs = {
    'name': 'wkconnect',
    'version': '21.8.0',
    'description': '',
    'long_description': '# webknossos-connect\nA [webKnossos](https://github.com/scalableminds/webknossos) compatible data connector written in Python.\n\nwebKnossos-connect serves as an adapter between the webKnossos data store interface and other alternative data storage servers (e.g BossDB) or static files hosted on Cloud Storage (e.g. Neuroglancer Precomputed)\n\n[![Github Actions](https://github.com/scalableminds/webknossos-connect/actions/workflows/main.yml/badge.svg)](https://github.com/scalableminds/webknossos-connect/actions)\n\nAvailable Adapaters / Supported Data Formats:\n- [BossDB](https://bossdb.org/)\n- [Neuroglancer Precomputed](https://github.com/google/neuroglancer/tree/master/src/neuroglancer/datasource/precomputed)\n- [WKW](https://github.com/scalableminds/webknossos-wrap)\n- Tiled TIFF\n\n## Usage\n### 1. Installation / Docker\nInstall webKnossos-connect using Docker or use the instructions for native installation below.\n`docker-compose up --build webknossos-connect`\n\n### 2. Connecting to webKnossos\nRegister your webknossos-connect instance with your main webKnossos instance. Modify the webKnossos Postgres database:\n  ```\n  INSERT INTO "webknossos"."datastores"("name","url","publicurl","key","isscratch","isdeleted","isforeign","isconnector")\n  VALUES (E\'connect\', E\'http://localhost:8000\', E\'http://localhost:8000\', E\'secret-key\', FALSE, FALSE, FALSE, TRUE);\n  ```\n### 3. Adding Datasets\nAdd and configure datasets to webKnossos-connect to make them available for viewing in webKnossos\n\n#### 3.1 REST API\nYou can add new datasets to webKnossos-connect through the REST interface. POST a JSON configuration to:\n```\nhttp://<webKnossos-connect>/data/datasets?token\n```\nThe access `token` can be obained from your user profile in the webKnossos main instance. [Read more in the webKnosssos docs.](https://docs.webknossos.org/reference/rest_api#authentication)\n\nExample JSON body. More examples can be found [here](https://github.com/scalableminds/webknossos-connect/blob/master/data/datasets.json).\n```\n{\n    "boss": {\n        "Test Organisation": {\n            "ara": {\n                "domain": "https://api.boss.neurodata.io",\n                "collection": "ara_2016",\n                "experiment": "sagittal_50um",\n                "username": "<NEURODATA_IO_USER>",\n                "password": "<NEURODATA_IO_PW>"\n            },\n        }\n    },\n    "neuroglancer": {\n        "Test Organisation": {\n            "fafb_v14": {\n                "layers": {\n                    "image": {\n                        "source": "gs://neuroglancer-fafb-data/fafb_v14/fafb_v14_clahe",\n                        "type": "image"\n                    }\n                }\n            }\n        }\n    },\n    "tiff": {\n        "Test Organization": {\n            "my_2d_tiff_dataset": {\n                "scale": [2.1,2.1]\n            }\n        }\n    }\n}\n```\n\nNote that tiff datasets are hosted locally. Create compatible tifs with `vips tiffsave source.tif color.tif --tile --pyramid --bigtiff --compression none --tile-width 256 --tile-height 256` and save the generated `color.tif` file at `data/binary/sample_organization/my_2d_tiff_dataset`.\n\nCURL Example\n```\ncurl http:/<webKnossos-connect>/data/datasets -X POST -H "Content-Type: application/json" --data-binary "@datasets.json"\n```\n\n#### 3.2 webKnossos UI\nAlternatively, new datasets can be added directly through the webKnossos UI. Configure and import a new datasets from the webKnossos dashboard. (Dashboard -> Datasets -> Upload Dataset -> Add wk-connect Dataset)\n\n[Read more in the webKnossos docs.](https://docs.webknossos.org/guides/datasets#uploading-through-the-web-browser)\n\n#### 3.3 Default test datasets\n\nBy default, some public datasets are added to webKnossos-connect to get you started when using the Docker image.\n\n## Development\n### In Docker :whale:\n\n* Start it with `docker-compose up dev`\n* Run other commands `docker-compose run --rm dev pipenv run lint`\n* [Check below](#moar) for moar commands.\n* If you change the packages, rebuild the image with `docker-compose build dev`\n\n### Native\n#### Installation\n\nYou need Python 3.8 with `poetry` installed.\n\n```bash\npip install poetry\npoetry install\n```\n\n#### Run\n\n* Add webknossos-connect to the webKnossos database:\n  ```\n  INSERT INTO "webknossos"."datastores"("name","url","publicurl","key","isscratch","isdeleted","isforeign","isconnector")\n  VALUES (E\'connect\', E\'http://localhost:8000\', E\'http://localhost:8000\', E\'secret-key\', FALSE, FALSE, FALSE, TRUE);\n  ```\n* `python -m wkconnect`\n* ```\n  curl http://localhost:8000/api/neuroglancer/Demo_Lab/test \\\n    -X POST -H "Content-Type: application/json" \\\n    --data-binary "@datasets.json"\n  ```\n\n### Moar\n\nUseful commands:\n\n* Lint with `pylint` & `flake8`\n* Format with `black`, `isort` & `autoflake`\n* Type-check with `mypy`\n* Benchark with `timeit`\n* Trace with `py-spy`\n\nUse the commands:\n\n* `scripts/pretty.sh`\n* `scripts/pretty-check.sh`\n* `scripts/lint.sh`\n* `scripts/type-check.sh`\n* `benchmarks/run_all.sh`\n\nTrace the server on http://localhost:8000/trace.\n\n## License\nAGPLv3\nCopyright [scalable minds](https://scalableminds.com)\n',
    'author': 'scalable minds',
    'author_email': 'hello@scalableminds.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, List, NamedTuple, Optional, cast

import numpy as np

from ..utils.types import JSON, Box3D, Vec3D, Vec3Df


class DataStoreStatus(NamedTuple):
    ok: bool
    url: str


class DataRequest(NamedTuple):
    position: Vec3D
    zoomStep: int
    cubeSize: int
    fourBit: bool


class BoundingBox(NamedTuple):
    topLeft: Vec3D
    width: int
    height: int
    depth: int

    def box(self) -> Box3D:
        return Box3D.from_size(self.topLeft, Vec3D(self.width, self.height, self.depth))

    @classmethod
    def from_box(cls, box: Box3D) -> BoundingBox:
        return cls(box.left, *box.size())


class DataSourceId(NamedTuple):
    team: str
    name: str


class Histogram(NamedTuple):
    elementCounts: List[int]
    numberOfElements: int
    min: float
    max: float


@dataclass(unsafe_hash=True)
class DataLayer:
    name: str
    category: str
    boundingBox: BoundingBox
    resolutions: List[Vec3D]
    elementClass: str
    dataFormat: str = field(default="wkw", init=False)
    wkwResolutions: JSON = field(init=False)
    largestSegmentId: Optional[int] = None
    mappings: Optional[List[Any]] = field(default=None, init=False)

    def __post_init__(self) -> None:
        assert self.category in ["color", "segmentation"]
        assert self.elementClass in ["uint8", "uint16", "uint24", "uint32", "uint64"]
        if self.category == "segmentation":
            if self.largestSegmentId is None:
                self.largestSegmentId = cast(int, np.iinfo(self.elementClass).max)
            self.mappings = []
        self.wkwResolutions = [
            {"resolution": i, "cubeLength": 1024} for i in self.resolutions
        ]


class DataSource(NamedTuple):
    id: DataSourceId
    dataLayers: List[DataLayer]
    scale: Vec3Df


class UnusableDataSource(NamedTuple):
    id: DataSourceId
    status: str



import asyncio
import json
import os
import tempfile
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, cast

import compressed_segmentation
import jpeg4py as jpeg
import numpy as np
from aiohttp import ClientResponseError, ClientSession
from async_lru import alru_cache
from gcloud.aio.auth import Token

from ...utils.exceptions import RuntimeErrorWithUserMessage
from ...utils.json import from_json
from ...utils.types import JSON, Box3D, Vec3D, Vec3Df
from ..backend import Backend, DatasetInfo
from .models import Dataset, Layer

DecoderFn = Callable[[bytes, str, Vec3D, Optional[Vec3D]], np.ndarray]
Chunk = Tuple[Box3D, np.ndarray]


class NeuroglancerAuthenticationError(RuntimeError):
    pass


class NeuroglancerAuthenticationMissingError(RuntimeErrorWithUserMessage):
    pass


async def get_header(token: Optional[Token]) -> Dict[str, str]:
    if token is None:
        return {}
    else:
        try:
            token_str = await token.get()
        except Exception as e:
            raise NeuroglancerAuthenticationError(*e.args)
        if token_str is None:
            return {}
        return {"Authorization": "Bearer " + token_str}


class Neuroglancer(Backend):
    def __init__(self, config: Dict, http_client: ClientSession) -> None:
        self.http_client = http_client
        self.decoders: Dict[str, DecoderFn] = {
            "raw": self.__decode_raw,
            "jpeg": self.__decode_jpeg,
            "compressed_segmentation": self.__decode_compressed_segmentation,
        }

    def create_token(self, credentials: Optional[JSON]) -> Optional[Token]:
        if credentials is None:
            return None
        tmpfile = tempfile.NamedTemporaryFile(mode="w+", delete=False)
        json.dump(credentials, tmpfile)
        tmpfile.close()
        try:
            token = Token(
                service_file=tmpfile.name,
                session=self.http_client,
                scopes=["https://www.googleapis.com/auth/cloud-platform"],
            )
        finally:
            os.remove(tmpfile.name)
        return token

    async def __handle_layer(
        self, layer_name: str, layer: Dict[str, Any], token: Optional[Token]
    ) -> Tuple[str, Layer]:
        layer["source"] = (
            layer["source"]
            .replace("gs://", "https://storage.googleapis.com/")
            .replace("s3://", "https://s3.amazonaws.com/")
        )
        info_url = layer["source"] + "/info"

        try:
            async with await self.http_client.get(
                info_url, headers=await get_header(token)
            ) as r:
                response = await r.json(content_type=None)
        except ClientResponseError as e:
            if e.status == 403:
                raise NeuroglancerAuthenticationMissingError(
                    "Could not authenticate the neuroglancer dataset, "
                    + "please add authentication"
                )
            else:
                raise e
        layer.update(response)
        layer["scales"] = sorted(layer["scales"], key=lambda scale: scale["resolution"])
        min_resolution = Vec3Df(*layer["scales"][0]["resolution"])
        for scale in layer["scales"]:
            resolution = Vec3Df(*scale["resolution"])
            assert resolution % min_resolution == Vec3D.zeros()
            scale["resolution"] = (resolution // min_resolution).to_int()
            scale["voxel_offset"] = Vec3D(*scale["voxel_offset"]).to_int()
        layer["relative_scale"] = min_resolution.to_int()
        return (layer_name, from_json(layer, Layer))

    async def handle_new_dataset(
        self, organization_name: str, dataset_name: str, dataset_info: JSON
    ) -> DatasetInfo:
        try:
            token = self.create_token(dataset_info.get("credentials", None))
        except Exception as e:
            NeuroglancerAuthenticationError(*e.args)
        layers = dict(
            await asyncio.gather(
                *(
                    self.__handle_layer(layer_name, layer, token)
                    for layer_name, layer in dataset_info["layers"].items()
                )
            )
        )
        dataset = Dataset(organization_name, dataset_name, layers, token)
        return dataset

    def __decode_raw(
        self, buffer: bytes, data_type: str, chunk_size: Vec3D, _: Optional[Vec3D]
    ) -> np.ndarray:
        return np.frombuffer(buffer, dtype=data_type).reshape(chunk_size, order="F")

    def __decode_jpeg(
        self, buffer: bytes, data_type: str, chunk_size: Vec3D, _: Optional[Vec3D]
    ) -> np.ndarray:
        np_bytes = np.frombuffer(buffer, dtype="uint8")
        return (
            jpeg.JPEG(np_bytes)
            .decode()[:, :, 0]
            .astype(data_type)
            .T.reshape(chunk_size, order="F")
        )

    def __decode_compressed_segmentation(
        self,
        buffer: bytes,
        data_type: str,
        chunk_size: Vec3D,
        block_size: Optional[Vec3D],
    ) -> np.ndarray:
        assert block_size is not None
        return compressed_segmentation.decompress(
            buffer, chunk_size, data_type, block_size, order="F"
        )

    def __chunks(
        self, requested: Box3D, ds_frame: Box3D, wk_chunk_size: Vec3D
    ) -> Iterable[Box3D]:
        # clip data outside available data
        inside = requested.intersect(ds_frame)

        # align request to chunks
        aligned = (inside - ds_frame.left).div(
            wk_chunk_size
        ) * wk_chunk_size + ds_frame.left

        for chunk_offset in aligned.range(offset=wk_chunk_size):
            chunk = Box3D.from_size(chunk_offset, wk_chunk_size)
            # chunk is capped to fit the dataset:
            yield chunk.intersect(ds_frame)

    @alru_cache(maxsize=2 ** 12, cache_exceptions=False)
    async def __read_chunk(
        self,
        layer: Layer,
        scale_key: str,
        chunk_box: Box3D,
        decoder_fn: DecoderFn,
        compressed_segmentation_block_size: Optional[Vec3D],
        token: Optional[Token],
    ) -> Chunk:
        url_coords = "_".join(
            f"{left_dim}-{right_dim}" for left_dim, right_dim in zip(*chunk_box)
        )
        data_url = f"{layer.source}/{scale_key}/{url_coords}"

        async with await self.http_client.get(
            data_url, headers=await get_header(token)
        ) as r:
            response_buffer = await r.read()
        chunk_data = decoder_fn(
            response_buffer,
            layer.data_type,
            chunk_box.size(),
            compressed_segmentation_block_size,
        ).astype(layer.wk_data_type())
        return (chunk_box, chunk_data)

    def __cutout(self, chunks: List[Chunk], box: Box3D) -> np.ndarray:
        result = np.zeros(box.size(), dtype=chunks[0][1].dtype, order="F")
        for chunk_box, chunk_data in chunks:
            inner = chunk_box.intersect(box)
            result[(inner - box.left).np_slice()] = chunk_data[
                (inner - chunk_box.left).np_slice()
            ]
        return result

    async def read_data(
        self,
        abstract_dataset: DatasetInfo,
        layer_name: str,
        zoom_step: int,
        wk_offset: Vec3D,
        shape: Vec3D,
    ) -> Optional[np.ndarray]:
        dataset = cast(Dataset, abstract_dataset)
        layer = dataset.layers[layer_name]
        # we can use zoom_step as the index, as the scales are sorted
        # see assertion in the Layer initialization
        scale = layer.scales[zoom_step]
        decoder = self.decoders[scale.encoding]

        # convert to coordinate system of current scale
        offset = wk_offset // scale.resolution
        box = Box3D.from_size(offset, shape)

        chunk_boxes = self.__chunks(box, scale.box(), scale.chunk_sizes[0])
        try:
            chunks = await asyncio.gather(
                *(
                    self.__read_chunk(
                        layer,
                        scale.key,
                        chunk_box,
                        decoder,
                        scale.compressed_segmentation_block_size,
                        dataset.token,
                    )
                    for chunk_box in chunk_boxes
                )
            )
        except ClientResponseError:
            # will be reported in MISSING-BUCKETS, frontend will retry
            return None

        return self.__cutout(chunks, box)

    def clear_dataset_cache(self, dataset: DatasetInfo) -> None:
        self.__read_chunk.cache_clear()  # pylint: disable=no-member



import asyncio
import json
import logging
import os
from copy import deepcopy
from types import SimpleNamespace
from typing import Any, Awaitable, Callable, Dict, List, Tuple, Type

from aiohttp import ClientSession
from aiohttp.client_exceptions import ClientConnectorError, ClientResponseError
from sanic import Sanic, response
from sanic.exceptions import SanicException
from sanic.handlers import ErrorHandler
from sanic.request import Request
from sanic_cors import CORS
from uvloop import Loop

from .backends.backend import Backend
from .backends.boss.backend import Boss
from .backends.neuroglancer.backend import Neuroglancer
from .backends.tiff.backend import Tiff
from .backends.wkw.backend import Wkw
from .repository import Repository
from .routes import routes
from .utils.exceptions import exception_traceback, format_exception
from .utils.scheduler import repeat_every_seconds
from .utils.types import JSON
from .webknossos.client import WebKnossosClient as WebKnossos
from .webknossos.models import DataSource, DataSourceId, UnusableDataSource

logger = logging.getLogger()
available_backends: List[Type[Backend]] = [Boss, Neuroglancer, Tiff, Wkw]


class ServerContext(SimpleNamespace):
    http_client: ClientSession
    repository: Repository
    webknossos: WebKnossos
    backends: Dict[str, Backend]


class Server(Sanic):
    ctx: ServerContext

    def __init__(self) -> None:
        super().__init__("wkconnect")

    async def add_dataset(
        self,
        dataset_config: JSON,
        backend_name: str,
        organization_name: str,
        dataset_name: str,
        report_to_wk: bool = True,
    ) -> DataSource:
        backend = self.ctx.backends[backend_name]
        dataset = await backend.handle_new_dataset(
            organization_name, dataset_name, deepcopy(dataset_config)
        )
        self.ctx.repository.add_dataset(backend_name, dataset)
        wk_dataset = dataset.to_webknossos()
        if report_to_wk:
            await self.ctx.webknossos.report_dataset(wk_dataset)
        return wk_dataset

    async def load_persisted_datasets(self) -> None:
        def expandvars_hook(dict: Dict[str, Any]) -> Dict[str, Any]:
            for key, val in dict.items():
                if isinstance(val, str):
                    dict[key] = os.path.expandvars(val)
            return dict

        try:
            with open(self.config["datasets_path"]) as datasets_file:
                datasets = json.load(datasets_file, object_hook=expandvars_hook)
        except FileNotFoundError:
            datasets = {}
        dataset_tuples = [
            (backend, organization, dataset, dataset_details)
            for backend, backend_details in datasets.items()
            for organization, organization_details in backend_details.items()
            for dataset, dataset_details in organization_details.items()
        ]
        results = await asyncio.gather(
            *(
                self.add_dataset(
                    dataset_config=dataset_details,
                    backend_name=backend,
                    organization_name=organization,
                    dataset_name=dataset,
                    report_to_wk=False,
                )
                for backend, organization, dataset, dataset_details in dataset_tuples
            ),
            return_exceptions=True,
        )
        for result in results:
            if isinstance(result, Exception):
                logger.error(exception_traceback(result))

        usable = [result for result in results if not isinstance(result, Exception)]
        unusable = [
            UnusableDataSource(
                DataSourceId(organization, dataset), status=format_exception(result)
            )
            for result, (_, organization, dataset, _) in zip(results, dataset_tuples)
            if isinstance(result, Exception)
        ]
        await self.ctx.webknossos.report_all_datasets(usable + unusable)


app = Server()
CORS(app, automatic_options=True)

with open("data/config.json") as config_file:
    app.config.update(json.load(config_file))


## ERROR HANDLING


class CustomErrorHandler(ErrorHandler):
    def default(self, request: Request, exception: Exception) -> response.HTTPResponse:
        """handles errors that have no other error handlers assigned"""
        if isinstance(exception, SanicException):
            return super().default(request, exception)
        else:
            message = format_exception(exception) + "."
            stack = exception_traceback(exception)
            message_json = {"messages": [{"error": message, "chain": [stack]}]}
            logger.error(stack)
            return response.json(message_json, status=500)


app.error_handler = CustomErrorHandler()


## TASKS ##


@app.listener("before_server_start")
async def setup(app: Server, loop: Loop) -> None:
    def instanciate_backend(backend_class: Type[Backend]) -> Tuple[str, Backend]:
        backend_name = backend_class.name()
        config = app.config["backends"].get(backend_name, {})
        return (backend_name, backend_class(config, app.ctx.http_client))

    app.ctx.http_client = await ClientSession(raise_for_status=True).__aenter__()
    app.ctx.repository = Repository()
    app.ctx.webknossos = WebKnossos(app.config, app.ctx.http_client)
    app.ctx.backends = dict(map(instanciate_backend, available_backends))


@app.listener("before_server_stop")
async def stop_tasks(app: Server, loop: Loop) -> None:
    for task in asyncio.Task.all_tasks():
        if not task == asyncio.Task.current_task():
            task.cancel()


@app.listener("after_server_stop")
async def shutdown(app: Server, loop: Loop) -> None:
    await asyncio.gather(
        *(backend.on_shutdown() for backend in app.ctx.backends.values())
    )
    await app.ctx.http_client.__aexit__(None, None, None)


## ROUTES ##


app.blueprint(routes)


@app.route("/data/health")
async def health(request: Request) -> response.HTTPResponse:
    return response.text("Ok")


@app.route("/api/buildinfo")
async def build_info(request: Request) -> response.HTTPResponse:
    return response.json(
        {
            "webknossosDatastore": {
                "webknossos-connect": {
                    "name": "webknossos-connect",
                    "version": "0.1",
                    "datastoreApiVersion": "1.0",
                }
            }
        }
    )


## MAIN ##


if __name__ == "__main__":

    def add_regular_interaction(
        action_name: str, fn: Callable[[Server], Awaitable[None]]
    ) -> None:
        @repeat_every_seconds(60)
        async def do_interaction(app: Server) -> None:
            try:
                await fn(app)
            except ClientConnectorError:
                logger.warning(f"Could not {action_name}, retrying in 1 min.")
            except ClientResponseError as exception:
                logger.warning(
                    f"Could not {action_name}, retrying in 1 min. Got {format_exception(exception)}."
                )

        app.add_task(repeat_every_seconds(10 * 60)(do_interaction))

    add_regular_interaction(
        "ping webknossos", lambda app: app.ctx.webknossos.report_status()
    )
    add_regular_interaction(
        "report datasets to webknossos", lambda app: app.load_persisted_datasets()
    )

    # We need to run sanic in the regular asyncio event loop
    loop = asyncio.get_event_loop()
    server_coroutine = app.create_server(
        host=app.config["server"]["host"],
        port=app.config["server"]["port"],
        access_log=False,
        return_asyncio_server=True,
        asyncio_server_kwargs=None,
    )
    server_task = asyncio.ensure_future(server_coroutine, loop=loop)
    server = loop.run_until_complete(server_task)
    if server:
        server.after_start()
        try:
            loop.run_forever()
        except KeyboardInterrupt:
            server.before_stop()
            loop.stop()
        finally:
            server.after_stop()
